#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>
#include <inttypes.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "debug.h"
#include "libdaemon.h"
#include "libcli.h"

#define LIBDMN_DEBUG

#define ERR(format, ...)  libdmn_log(h, 0, "ERR: "format"\n", ##__VA_ARGS__)
#define WARN(format, ...) libdmn_log(h, 1, "WARN: "format"\n", ##__VA_ARGS__)
#define MSG(format, ...)  libdmn_log(h, 2, format"\n", ##__VA_ARGS__)
#define LOG(format, ...)  MSG(format, ##__VA_ARGS__)

#define ERR2(format, ...)  libdmn_log2(d, 0, "ERR: "format"\n", ##__VA_ARGS__)
#define WARN2(format, ...) libdmn_log2(d, 1, "WARN: "format"\n", ##__VA_ARGS__)
#define MSG2(format, ...)  libdmn_log2(d, 2, format"\n", ##__VA_ARGS__)
#define LOG2(format, ...)  MSG2(format, ##__VA_ARGS__)


#ifdef LIBDMN_DEBUG
#define DBGL(level, format, ...) libdmn_log(h, 3+level, "DBG: %s: "format"\n", __FUNCTION__, ##__VA_ARGS__)
#define DBG(format, ...) DBGL(0, format, ##__VA_ARGS__)
#define DBGL2(level, format, ...) libdmn_log2(d, 3+level, "DBG: %s: "format"\n", __FUNCTION__, ##__VA_ARGS__)
#define DBG2(format, ...) DBGL2(0, format, ##__VA_ARGS__)
#else
#define DBGL(level, format, ...)
#define DBG(format, ...)
#define DBGL2(level, format, ...)
#define DBG2(format, ...)
#endif

#define MKRESPONSE(rhdr, ln, cod, sts) \
  rhdr->status = sts; \
  rhdr->hdr.code = htons(cod | LIBDMN_MSG_RESPONSE); \
  rhdr->hdr.len = htons(ln);

#define LIBDMN_CMEXE_TIMEOUT 1000

#define DM_DATA(d) d->ud;


int libdmn_handle_args(daemon_t *d, int argc, char **argv);

int libdmn_set_timeval_ms(struct timeval *tv, uint32_t msec)
{
  memset(tv, 0 ,sizeof(struct timeval));
  tv->tv_sec = msec/1000;
  tv->tv_usec = (msec - (msec/1000)*1000)*1000;
  return 0;
}

uint64_t libdmn_gettime_ms()
{
  uint64_t t;
  struct timeval tv;
  memset(&tv, 0 ,sizeof(struct timeval));
  gettimeofday(&tv, 0);
  t = tv.tv_sec*1000+tv.tv_usec/1000;
  return t;
}

int libdmn_waitfd(libdmn_t *h, int fd, uint16_t timer, char m)
{
  fd_set set;
  struct timeval tv;
  int r;

  FD_ZERO(&set);
  FD_SET(fd, &set);

  libdmn_set_timeval_ms(&tv, timer);

  while (1) {

    DBGL(4, "wait '%c' timer %u", m, timer);
    r = select(fd+1, (m=='r')?&set:0, (m=='w')?&set:0, 0, &tv);

    if (r<0) {

      if (errno == EINTR) {
        if (timer)
          return 1;
        else
          continue;
      }

      ERR("select %s", strerror(errno));
      return -1;
    }

    if (r==0)
      return 1;

    if (r>0)
      break;
  }

  return 0;
}



int fdlist_find(fdlist_t *fdl, int fd)
{
  uint16_t i;

  for (i=0; i<fdl->cnt; ++i) {
    if ( ((fdlist_item_t *)array_at(fdl->a, i))->fd == fd )
      return i;
  }
  return -1;
}

int fdlist_add(fdlist_t *fdl, int fd, uint8_t mode)
{
  fdlist_item_t it;
  it.fd = fd;
  it.mode = mode;
  return list_add(fdl, &it);
}

int fdlist_del(fdlist_t *fdl, int fd)
{
  int i;
  i = fdlist_find(fdl, fd);  
  if (i<0)
    return 1;
  return list_del_by_idx(fdl, i);
}

int libdmn_save_read(libdmn_t *h, int fd, void *buf, size_t size)
{
  size_t rcv=0;
  int r;

    while (1) {

      r=libdmn_waitfd(h, fd, 100, 'r');

      if (r) {
        if (r>0)
          ERR("io timeout");
        return -1;
      }

      r=read(fd, &((char*)buf)[rcv], size-rcv);

      if (r<0) {
        if (r!=EINTR) {
          ERR("read %s", strerror(errno));
          return -1;
        }
      }

      rcv += r;
      if (rcv>=size)
        break ;

    }

   return 0;
}

int libdmn_save_write(libdmn_t *h, int fd, void *buf, size_t size)
{
  return write(fd, buf, size);
}

int libdmn_read_message(libdmn_t *h, int fd,  uint8_t *buf)
{  
  libdmn_msg_hdr_t *rhdr=(libdmn_msg_hdr_t *)buf;

  DBG("fd %i", fd);

  if (libdmn_save_read(h, fd, buf, sizeof(libdmn_msg_hdr_t)))
    return -1;

  DBGL(4, "msg: len: %u code: %04X", ntohs(rhdr->len), ntohs(rhdr->code));

  if (libdmn_save_read(h, fd, &buf[sizeof(libdmn_msg_hdr_t)], ntohs(rhdr->len)))
    return -1;

  return 0;
}

int libdmn_write_message(libdmn_t *h, int fd,  uint8_t *buf)
{
  libdmn_msg_hdr_t *whdr=(libdmn_msg_hdr_t *)buf;
  DBG("write_msg: %i len: %i", fd, sizeof(libdmn_msg_hdr_t) + ntohs(whdr->len));
  return libdmn_save_write(h, fd, buf, sizeof(libdmn_msg_hdr_t) + ntohs(whdr->len));
}

int libdmn_strtook(char **start, char **end, char **next)
{
  char state=0, *c;
  int ch;

  if (!(*start))
    return 1;

  for (c=(*start); c!=0; ++c) {

    ch = (*c);

    switch (state) {
      case 0:
       if (isspace((int)(*c)))
         continue;
       if (ch=='"') {
         state=1;
         (*start)=(c+1);
       } else {
         state=2;
         (*start)=c;
       }
      break;
      case 1:
       if (ch=='"') {
         (*c)=0;
         (*end) = (c-1);
         (*next) = (c+1);
         if ((*start) >= (*end))
           return 1;
         return 0;
       }
      break;
      case 2:
       if (isspace((int)(*c))) {
         (*c)=0;
         (*end) = c;
         (*next) = (c+1);
         if ((*start) >= (*end))
           return 1;
         return 0;
       }
       break;
      default:
       return 1; 
    }

  }

  if (state==2) {
    (*end) = (c-1);
    (*next) = 0;
    if ((*start) < (*end))
        return 0;
  }

  return 1;
}

int libdmn_process_str_cmd(libdmn_t *h, void *rbuf, libdmn_str_cmd_tbl_t *cmd_tbl, uint16_t cmd_count, void *data)
{
  int i;
  char *argv[32], *p;
  int argc = 0;

  /* build argc argv */

  /*
  while (!libdmn_strtook(&s, &e, &n)) {
    (*e)=0;
    argv[argc++]=s;
    s=n;
  }
  */

  if (!cmd_count)
    return 1;

  p = strtok(((libdmn_msg_str_cmd_t *)rbuf)->cmd, " ");

  while (p && argc < 32) {
    argv[argc++] = p;
    p = strtok(0, " ");
  }

  /*  */
  DBG("argc %i", argc);

  if (argc < 1)
    return -1;

  for (i=0; i<argc; ++i) {
    DBG("argv[%i] %s", i, argv[i]);
  }

  /* find command */
  for (i = 0;  i < cmd_count; ++i)
    if (!strcmp(argv[0], cmd_tbl[i].cmd))
      break;

  if (i >= cmd_count)
    return -2;

  DBG("handler found %u", i);

  return cmd_tbl[i].handler(argc, argv, data);

}

int libdmn_log2(daemon_t *d, int mlevel, const char *format, ...)
{
  char buf[128];
  va_list args;
  va_start( args, format);
  vsnprintf(buf, 128, format, args);
  va_end( args );

  log_print(d->log, mlevel, "%s", buf);
  return 0;
}

int libdmn_process_str_cmd2(daemon_t *d)
{
  daemon_common_var_t *v = &d->cv;

  int i, r=0 ;
  char *argv[32], *p;
  int argc = 0;

  if (!list_count(v->cmd)) {
    r=1;
    goto error;
  }

  p = strtok(((libdmn_msg_str_cmd_t *)v->buf)->cmd, " ");

  while (p && argc < 32) {
    argv[argc++] = p;
    p = strtok(0, " ");
  }

  /*  */
  DBG2("argc %i", argc);

  if (argc < 1) {
    r=1;
    goto error;
  }

  for (i=0; i<argc; ++i)
    DBG2("argv[%i] %s", i, argv[i]);

  /* find command */
  for (i = 0;  i < list_count(v->cmd); ++i)
    if (!strcmp(argv[0], ((libdmn_str_cmd_tbl_t*)list_at(v->cmd, i))->cmd))
      break;

  if (i >= list_count(v->cmd)) {
    r=2;
    goto error;
  }

  DBG2("handler found %u", i);

  return ((libdmn_str_cmd_tbl_t*)list_at(v->cmd, i))->handler(argc, argv, d);

error:

  if (r)
    libdmn_mk_str_cmd_response(d, 0x0F, (r==2)?"Command not found!":"ERR %i", r);

  return r;
}


/*
int libdmn_process_message(int fd,  uint8_t *rbuf, uint8_t *wbuf, libdmn_str_cmd_tbl_t *cmd_tbl, uint16_t cmd_count)
{
  libdmn_msg_hdr_t *rhdr = (libdmn_msg_hdr_t *)rbuf;
  uint16_t code = ntohs(rhdr->code);
  int r;

  DBG("Process message len: %u code: %04X", ntohs(rhdr->len), code);
  switch (code) {
    case LIBDMN_MSG_STR_CMD:

    break;
  }

  if (r) {
    MKRESPONSE( ((libdmn_msg_response_hdr_t*)wbuf),
    0, code, 0x0F);
  }

  return 0;
}
*/

int libdmn_execute_cm(libdmn_t *h, char *path, void *buf, size_t bufsize, char *command)
{
  int r, fd;
  struct sockaddr_un sa;

  DBG("'%s'", command);

  strcpy(sa.sun_path, path);
  sa.sun_family = AF_UNIX;

  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (fd == -1) {
    DBG("socket(): %s", strerror(errno));
    goto error;
  }

  if (connect(fd, (struct sockaddr*)&sa, sizeof(sa))) {
    DBG("Can't connect to '%s': %s", path, strerror(errno));
    goto error;
  }

  libdmn_msg_str_cmd_t *hdr=(libdmn_msg_str_cmd_t *)buf;
  hdr->hdr.code = htons(LIBDMN_MSG_STR_CMD);
  hdr->hdr.len = htons(strlen(command)+1);
  strncpy(hdr->cmd, command, 64);
  libdmn_write_message(h, fd, buf);

  r = libdmn_waitfd(h, fd, LIBDMN_CMEXE_TIMEOUT, 'r');

  if (r) {
    if (r>0)
      ERR("timeout");
    goto error;
  }

  libdmn_read_message(h, fd, buf);
  libdmn_msg_str_cmd_r_t *hdr_r = (libdmn_msg_str_cmd_r_t *)buf;

  DBG("response status %02X", hdr_r->rhdr.status);

  fprintf(stdout, "%s\n", hdr_r->response);

  if (hdr_r->rhdr.status!=0) {
    goto error;
  }

  shutdown(fd, SHUT_RDWR);
  close(fd);

  return 0;

error:

  if (fd >= 0) {
   shutdown(fd, SHUT_RDWR);
   close(fd);
  }

  return 1;
}

int libdmn_handle_event(libdmn_t *h, libdmn_childs_t *ch, char *scriptfile, char *args)
{
  pid_t pid;
  char cmd[64];
  snprintf(cmd, sizeof(cmd), "%s %s", scriptfile, args);

  pid = fork();

  if (pid>0) {
    DBG("Child %i started\n", pid);
    ch->childs[ch->childs_cnt] = pid;
    ch->childs_cnt++;
    return 0;
  }

  if (pid<0) {
    /*DBG("Error creating process '%s'\n", strerror(errno));*/
    return -1;
  }

  /* Lock child to serialize event handling */
  /* this fd  & lock clears automaticaly on process exit */
  int fd = open(scriptfile, O_RDWR);
  if (fd>=0) {
    lseek(fd, 0, SEEK_SET);
    lockf(fd, F_LOCK, 0);
  }

  DBG("Executed '%s'", cmd);
  execl("/bin/sh", "sh", "-c", cmd, NULL);

  DBG("exec: %s", strerror(errno));

  return -1;
}

void libdmn_del_child(libdmn_t *h, libdmn_childs_t *ch, pid_t pid)
{
  int i;
  char f = 0;

  for (i=0;i<ch->childs_cnt;++i)
    if ( (pid == ch->childs[i]) || f) {
      f = 1;
      if ((unsigned int)i < (sizeof(ch->childs)/sizeof(pid_t)-1) )
        ch->childs[i]=ch->childs[i+1];
    }
  if (f)
    ch->childs_cnt--;
}

void libdmn_terminate_childs(libdmn_t *h, libdmn_childs_t *ch)
{
  int cnt,sts, i;

  for (i=0; i<ch->childs_cnt; ++i) {
    DBG("Send SIGTERM to %i", ch->childs[i]);
    kill(ch->childs[i], SIGTERM);
  }

  cnt=5;
  while ((ch->childs_cnt) && (!cnt)) {
    sleep(1);
    cnt--;
  }

  if (!ch->childs_cnt)
    return;

  ERR("%u childs not terminated correctly", ch->childs_cnt);

  for (i=0; i<ch->childs_cnt; ++i) {
    DBG("Send SIGKILL to %i", ch->childs[i]);
    kill(ch->childs[i], SIGKILL);
    waitpid(ch->childs[i], &sts, 0);
  }

}

int libdmn_setsighandler(libdmn_t *h, int sig, void (*sig_handler)(int))
{
  struct sigaction act;
  memset(&act, 0, sizeof(act));
  act.sa_handler = sig_handler;
  if (sigaction(sig, &act, NULL) == -1) {
   DBG("sigaction: %s\n", strerror(errno));
   return 1;
  }

  return 0;
}


int libdmn_init(libdmn_t **h, const log_opts_t *log_opts, const char *prfx)
{  
  (*h) = malloc(sizeof(libdmn_t));

  if (!(*h))
    return 1;

  (*h)->dbg = malloc(sizeof(log_desc_t));

  if (!((*h)->dbg))
    goto error;

  if (log_init((log_desc_t*)((*h)->dbg), log_opts, prfx))
    goto error;

  return 0;

error:
  libdmn_free((*h));
  *h = NULL;
  return 1;
}

int libdmn_free(libdmn_t *h)
{
  if (h) {
    if (h->dbg) {
      log_free(h->dbg);
      free( h->dbg );
      h->dbg = NULL;
    }
    free(h);
  }
  return 0;
}

int libdmn_log(libdmn_t *h, int mlevel, const char *format, ...)
{
  char buf[128];  
  va_list args;

  va_start( args, format);
  vsnprintf(buf, 128, format, args);
  va_end( args );

  if( h && h->dbg ) {
    log_print((log_desc_t*)h->dbg, mlevel, "%s", buf);
  }

  return 0;
}

array_t *array_create(size_t isize, uint16_t icnt)
{
      array_t *a=0;
      a = malloc(sizeof(array_t));
      a->isize = isize;
      a->maxcnt = icnt;
      a->head = malloc(isize*icnt);
      memset(a->head, 0, isize*icnt);
      return a;
}

void array_free(array_t *a)
{
  if (a) {
   free(a->head);
   free(a);
  }
}

void *array_at(array_t *a, uint16_t inum)
{
  if (inum>=a->maxcnt)
    return 0;
  return &(a->head[(a->isize)*inum]);
}

uint16_t array_idx(array_t *a, void *item)
{
  return ((uint8_t*)item - (a->head))/a->isize;
}

uint16_t array_size(array_t *a)
{
  return a->maxcnt;
}

size_t array_isize(array_t *a)
{
  return a->isize;
}


bitmap_t *create_bitmap(uint8_t bitsize, uint8_t cnt)
{
  bitmap_t *b=0;
  b = malloc(sizeof(bitmap_t));
  b->size = ((bitsize*cnt)>>3)+1;
  b->head = malloc(b->size);
  return b;
}

void free_bitmap(bitmap_t *b)
{
  if (b) {
    free(b->head);
    free(b);
    b=0;
  }
}

list_t *list_create(size_t isize, uint16_t maxcnt)
{
  list_t *l;
  l = malloc(sizeof(list_t));
  l->a = array_create(isize, maxcnt);
  l->cnt = 0;
  return l;
}

void list_free(list_t *l)
{
  if (l) {
    array_free(l->a);
    free(l);
  }
}


uint16_t list_count(list_t *l)
{
  return l->cnt;
}

uint16_t list_size(list_t *l)
{
  return l->a->maxcnt;
}

int list_add(list_t *l, void *item)
{
  void *p;
  p = array_at(l->a, l->cnt);  
  if (!p)
    return 1;
  memcpy(p, item, l->a->isize);
  l->cnt++;
  return 0;
}

int list_find(list_t *l, void *item)
{
  uint16_t i;
  if (!item)
    return -1;
  for (i=0; i<array_size(l->a); ++i) {     
     if (!memcmp(array_at(l->a, i), item, l->a->isize))
       return i;
  }
  return -1;
}

int list_del_by_idx(list_t *l, uint16_t i)
{
  if (i>=l->cnt)
    return 1;
  if ((l->cnt-i)>1)
    memcpy(array_at(l->a, i), array_at(l->a, i+1), sizeof(void *)*(l->cnt-i-1));
  l->cnt--;

  return 0;
}

int list_del(list_t *l, void *item)
{
  int i;
  i = list_find(l, item);
  if (i<0)
    return 1;
  return list_del_by_idx(l, i);
}

void *list_at(list_t *l, uint16_t idx)
{
  if (idx>=l->cnt)
    return 0;
  return array_at(l->a, idx);
}

void *list_last(list_t *l)
{
  if (!(l->cnt))
    return 0;
  return array_at(l->a, (l->cnt)-1);
}

void list_clear(list_t *l)
{
  l->cnt=0;
}

void ptrlist_free_with_items(list_t *l)
{
  uint16_t i;
  for (i=0; i<l->cnt; ++i)
    free((*((void **)list_at(l, i))));
  list_free(l);
}


uint8_t get_bitmap_item(bitmap_t *b, uint16_t inum)
{
  uint16_t sh,bn,sh2;
  uint8_t r;

  sh = (inum*(b->bitsize))&(0x7);
  bn = (inum*(b->bitsize))>>3;
  r = ((b->head[bn])>>sh);

  if ((sh+b->bitsize)>8) {
    sh2 = 8-((sh+(b->bitsize))-8);
    r |= (b->head[bn+1]) & (0xFF>>sh2);
  } else
    r &= (0xFF>>(b->bitsize));

  return r;
}

void set_bitmap_item(bitmap_t *b, uint16_t inum, uint8_t v)
{
}

uint8_t test_bitmap_item(bitmap_t *b)
{
  return 0;
}

int libdmn_wait_read_fds(daemon_t *d, fdlist_t *fdl, fdlist_t *fdr, uint16_t timer)
{
  fd_set set, rset;
  struct timeval tv;
  int r, maxfd, i;

  maxfd = -1;
  FD_ZERO(&set);
  DBGL2(2,"cnt: %u", fdlist_count(fdl));
  for (i=0; i<fdlist_count(fdl); ++i) {
    FD_SET(fdlist_at(fdl, i)->fd, &set);
    if (maxfd<fdlist_at(fdl, i)->fd)
      maxfd = fdlist_at(fdl, i)->fd;
  }

  libdmn_set_timeval_ms(&tv, timer);

  while (1) {

    rset = set;
    DBGL2(2, "wait 'r' timer %u", timer);
    r = select(maxfd+1, &rset, 0, 0, &tv);

    if (r<0) {

      if (errno==EINTR)  {
        if (timer)
          return 1;
        else
          continue;
      }

      ERR2("select %s", strerror(errno));
      return -1;
    }

    if (r==0)
      return 1;

    if (r>0)
      break;
  }

  fdlist_clear(fdr);
  for (i=0;i<fdlist_count(fdl);++i) {
     if (FD_ISSET(fdlist_at(fdl, i)->fd, &rset))
       fdlist_add(fdr, fdlist_at(fdl, i)->fd, 0);
  }

  return 0;
}



int libdmn_daemon_init_common_vars(daemon_t *d)
{
  daemon_common_var_t *v = &d->cv;

  memset(v, 0, sizeof(daemon_common_var_t));
  strncpy(v->name, d->get_name(d), sizeof(v->name));
  v->fd = -1;
  v->sleeptimer = 100; /* msec */
  v->bufsize = 2048;
  v->description = (d->get_description)?d->get_description():0;

  /* Socket to listen clients */
  snprintf(v->sa.sun_path, sizeof(v->sa.sun_path), "/tmp/%s.sock", v->name);
  v->sa.sun_family = AF_UNIX;

  return 0;
}


int libdmn_daemon_cleanup(daemon_t *d)
{
  daemon_common_var_t *v = &d->cv;

  if (d->cleanup)
    d->cleanup(d);

  if (v->fd>0)
    close(v->fd);
  if (d->v)
    free(d->v);
  if (v->buf)
    free(v->buf);

  if (v->fdl)
    fdlist_free(v->fdl);

  if (v->fdlr)
    fdlist_free(v->fdlr);

  if (v->cl_opts)
    cli_free_option_array2(v->cl_opts);

  if (v->cl_args)
    cli_free_option_array2(v->cl_args);

  if (v->mng_fdl)
    fdlist_free(v->mng_fdl);
  if (v->timers)
    list_free(v->timers);
  if (v->period_timers)
    ptrlist_free_with_items(v->period_timers);
  if (v->fdhandlers)
    list_free(v->fdhandlers);
  if (v->cmd)
    list_free(v->cmd);

  if (d->log) {
    log_free(d->log);
    free( d->log );
  }
    

  return 0;
}

int libdmn_daemon_exit(daemon_t *d, int code)
{
  if (code) {
    LOG2("Exit: with error %i", code);
  } else {
    LOG2("Exit: success");
  }
  libdmn_daemon_cleanup(d);
  exit(code);
}

void libdmn_daemon_sig_handler(daemon_t *d, int signum)
{
  LOG2("Signal received %i", signum);

  if (signum == SIGHUP) {
    DBG2("SIGHUP");
    if (d->reset)
        d->reset(d);
    return;
  }

  if ((signum == SIGINT) || (signum == SIGTERM)) {
    LOG2("Exiting");
    libdmn_daemon_exit(d, 0);
  }
}

int libdmn_process_message(daemon_t *d)
{
  daemon_common_var_t *v = &d->cv;
  libdmn_msg_hdr_t *rhdr = (libdmn_msg_hdr_t *)v->buf;
  uint16_t code = ntohs(rhdr->code);
  int r=0;

  DBG2("Process message len: %u code: %04X", ntohs(rhdr->len), code);
  switch (code) {
    case LIBDMN_MSG_STR_CMD:
      r = libdmn_process_str_cmd2(d);
      break;
    default:
      DBG2("Unknown msg: %04x", code);
      LIBDMN_MKRESPONSE(v->buf, 1, code, 0x0F);
      r = 0;
      break;
  }

  return r;
}

int libdmn_add_cm_handler(daemon_t *d, uint16_t cmid, libdmn_str_cmd_handler_t *handler)
{
  return 0;
}

int libdmn_add_str_cmd_handler(daemon_t *d, const char *name, libdmn_str_cmd_handler_t *handler)
{
  libdmn_str_cmd_tbl_t it;
  it.cmd = name;
  it.handler = *handler;
  list_add(d->cv.cmd, &it);
  return 0;
}

int libdmn_add_str_cmd_handlers(daemon_t *d, libdmn_str_cmd_tbl_t *table)
{
  while (table->cmd) {
    list_add(d->cv.cmd, table);
    ++table;
  }
  return 0;
}

void libdmn_mk_str_cmd_response(daemon_t *d, uint8_t statuscode, const char *fmt, ...)
{
  libdmn_msg_str_cmd_r_t *hdr = (libdmn_msg_str_cmd_r_t *)(d->cv.buf);
  va_list args;

  va_start(args, fmt);
  vsnprintf(hdr->response, d->cv.bufsize, fmt, args);
  va_end(args);

  LIBDMN_MKRESPONSE((&hdr->rhdr), strlen(hdr->response)+1,
             LIBDMN_MSG_STR_CMD, statuscode);

}

void libdmn_add_str_cmd_response_line(daemon_t *d, char **p, const char *fmt, ...)
{
  libdmn_msg_str_cmd_r_t *hdr = (libdmn_msg_str_cmd_r_t *)(d->cv.buf);
  va_list args;

  if ( (!p) || (!(*p)) )
    (*p) = hdr->response;

  va_start(args, fmt);
  vsnprintf((*p), d->cv.bufsize, fmt, args);
  va_end(args);

  (*p) += strlen(*p);
}

void libdmn_add_str_cmd_response_status(daemon_t *d, uint8_t statuscode)
{
  libdmn_msg_str_cmd_r_t *hdr = (libdmn_msg_str_cmd_r_t *)(d->cv.buf);

  LIBDMN_MKRESPONSE((&hdr->rhdr), strlen(hdr->response)+1,
             LIBDMN_MSG_STR_CMD, statuscode);
}

int libdmn_list_handler(int argc, char **argv, void *data)
{
  daemon_t *d = (daemon_t *)data;
  daemon_common_var_t *v = &d->cv;
  uint16_t i;
  char *p=0;

  libdmn_add_str_cmd_response_line(d, &p, "\n%s Supported commands:\n\n", v->name);

  for (i=0; i<list_count(v->cmd); ++i) {

    libdmn_str_cmd_tbl_t *it = list_at(v->cmd, i);
    libdmn_add_str_cmd_response_line(d, &p, "%-24s- %s\n",
                                     it->cmd, it->help);
  }

  libdmn_add_str_cmd_response_status(d, 0);
  return 0;
}

int libdmn_show_status_handler(int argc, char **argv, void *data)
{
  daemon_t *d = (daemon_t *)data;
  daemon_common_var_t *v = &d->cv;
  time_t t;
  uint64_t uptime;
  time(&t);
  uptime = t-v->starttime;
  libdmn_mk_str_cmd_response(d, 0, "uptime: %"PRIu64"s", uptime);
  return 0;
}

int libdmn_init_handler(int argc, char **argv, void *data)
{
  daemon_t *d = (daemon_t *)data;
  int r = 0;
  if (d->reset)
    r=d->reset(d);
  libdmn_mk_str_cmd_response(d, r, (r)?"ERR: %u":"OK", r);

  return r;
}

libdmn_str_cmd_tbl_t libdmn_str_cmd_table[] = {
  {"list", "Print list of commands", libdmn_list_handler},
  {"status", "Show runtime status", libdmn_show_status_handler},
  {"init", "Reinitialization", libdmn_init_handler},
  {"reset", "Reinitialization", libdmn_init_handler},
  {0, 0, 0}
};


libdmn_fd_handler_item_t* find_fd_handler_item(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  uint16_t i;
  libdmn_fd_handler_item_t *h;
  for (i=0; i<fdlist_count(v->fdhandlers); ++i) {
    h = ((libdmn_fd_handler_item_t *)list_at(v->fdhandlers, i));
    if ((h)&&(h->check(d, fd)))
      return h;
  }
  return 0;
}

int libdmn_daemon_add_fd(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  return fdlist_add(v->fdl, fd, 0);
}

int libdmn_daemon_del_fd(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  return fdlist_del(v->fdl, fd);
}

int libdmn_daemon_mng_init(daemon_t *d)
{
  daemon_common_var_t *v = &d->cv;
  return fdlist_add(v->fdl, v->fd, 0);
}

int libdmn_daemon_mng_check(daemon_t *d, int fd)
{
  DBG2("fd: %i", fd);
  if (d->cv.fd==fd)
    return 1;
  return 0;
}

int libdmn_daemon_mng_handle_close(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  int newfd;
  DBG2("ADD Client %i\n", fd);
  /* new connection */
  newfd = accept(v->fd, NULL, 0);
  fdlist_add(v->fdl, newfd, 0);
  fdlist_add(v->mng_fdl, newfd, 0);
  return 0;
}

int libdmn_daemon_mng_handle_rx(daemon_t *d, int fd)
{
//  daemon_common_var_t *v = &d->cv;
  DBG2(" %i!\n", fd);

  return 0;
}

int libdmn_daemon_mngcl_check(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  DBG2("%i cnt %u", fd, fdlist_count(v->mng_fdl));
  return (fdlist_find(v->mng_fdl, fd)>=0);
}

int libdmn_daemon_mngcl_handle_close(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  /* connection closed */
  DBG2("DEL client %i!", fd);
  fdlist_del(v->fdl, fd);
  fdlist_del(v->mng_fdl, fd);
  close(fd);
  return 0;
}

int libdmn_daemon_mngcl_handle_rx(daemon_t *d, int fd)
{
  daemon_common_var_t *v = &d->cv;
  DBG2("");

  if (libdmn_read_message(&d->h, fd, v->buf))
    return 1;

  libdmn_process_message(d);

  return libdmn_write_message(&d->h, fd, v->buf);
}

void libdmn_add_fd_handler(daemon_t *d, libdmn_fd_handler_item_t *i)
{
  list_add(d->cv.fdhandlers, i);
}

int libdmn_print_cmd_opts_help(daemon_t *d, char **p, cli_option_item_t **o)
{
  char s[33];
  while (*o) {

    switch ((*o)->type) {
      default:
      case CLI_OPTION_FLAG:
        snprintf(s, sizeof(s), "%s", (*o)->name);
      break;
      case CLI_OPTION_STR:
        snprintf(s, sizeof(s), "%s <string>", (*o)->name);
      break;
      case CLI_OPTION_UINT16: case CLI_OPTION_UINT32:
        snprintf(s, sizeof(s), "%s <num>", (*o)->name);
      break;
      case CLI_OPTION_UINT16X: case CLI_OPTION_UINT32X:
        snprintf(s, sizeof(s), "%s <hexnum>", (*o)->name);
      break;
    }
    if (!((*o)->flags&CLI_OPTIONFL_HIDDEN))
      libdmn_add_str_cmd_response_line(d, p, "-%-24s - %s\n", s, (*o)->help);
    o++;
  }
  return 0;
}

int libdmn_print_cmd_args_help(daemon_t *d, char **p, cli_option_item_t **o)
{
  while (*o) {
    libdmn_add_str_cmd_response_line(d, p, "[ %s ]  ", (*o)->help);
    o++;
  }
  libdmn_add_str_cmd_response_line(d, p, "\n");
  return 0;
}

int libdmn_handle_cmd_opts(daemon_t *d,
                           void *useroptions,
                           void *userargs,
                            int argc, char **argv)
{
  uint8_t help;
  int r=0;
  char *p=0;
  cli_option_item_t **options;
  cli_option_item_t **arguments;

  cli_option_item_t stdoptions[] = {
    {"h", "help", CLI_OPTION_FLAG, &help, 0, 0},
    {0, 0, 0, 0, 0, 0}
  };

  options = cli_create_option_array2(0);
  arguments = cli_create_option_array2(6);

  cli_add_options2(options, stdoptions);
  cli_add_options2(options, (cli_option_item_t*)useroptions);
  cli_add_options2(arguments, (cli_option_item_t*)userargs);

  int i=0;
  DBG2("argc %i", argc);
  while (i<argc) {
    DBG2("argv [%i]=%s", i, argv[i]);
    i++;
  }

  if (cli_handle_opt2(options, arguments, argc, argv)) {
    r=-1;
    goto exit;
  }

  if (help)  {

    if (arguments[0]) {
      libdmn_add_str_cmd_response_line(d, &p, "Arguments:\n");
      libdmn_print_cmd_args_help(d, &p, arguments);
      libdmn_add_str_cmd_response_line(d, &p, "\n");
    }

    if (options[0]) {
      libdmn_add_str_cmd_response_line(d, &p, "Options:\n");
      libdmn_print_cmd_opts_help(d, &p, options);
    }

    r=1;
    libdmn_add_str_cmd_response_status(d, 0);
    goto exit;
  }

exit:

  cli_free_option_array2(options);
  cli_free_option_array2(arguments);

  return r;
}

int libdmn_add_opts(daemon_t *d, cli_option_item_t *opts)
{
  daemon_common_var_t *v = &d->cv;

  return cli_add_options2(v->cl_opts, opts);
}

int libdmn_add_args(daemon_t *d, cli_option_item_t *args)
{
  daemon_common_var_t *v = &d->cv;

  return cli_add_options2(v->cl_args, args);
}

int libdmn_daemon_init(daemon_t *d, int argc, char **argv)
{
  daemon_common_var_t *v = &d->cv;
  libdmn_fd_handler_item_t hit;
  int r;

  cli_option_item_t opts[] = {
    {"h", "Print help", CLI_OPTION_FLAG, &v->help, 0, 0},
    {"d", "Log detalization level", CLI_OPTION_UINT16, &v->log_opts.loglevel, 0, 0},
    {"f", "Do not daemonize", CLI_OPTION_FLAG, &v->foreground, 0, 0},
    {"l", "Log file name or 'sys'", CLI_OPTION_STR, v->log_opts.logfile, sizeof(v->log_opts.logfile), 0},
    {"facility", "Log facility (for syslog only)", CLI_OPTION_STR, v->log_opts.facil_name, sizeof(v->log_opts.facil_name), 0},
    {"sleep-timer", "Wake up interval", CLI_OPTION_UINT16, &v->sleeptimer, 0, 0},
    {0, 0, 0, 0, 0, 0}
  };

  cli_option_item_t args[] = {
    {"", "Command", CLI_OPTION_STR, v->cm, sizeof(v->cm), 0 },
    {0, 0, 0, 0, 0, 0}
  };

  d->v = malloc(d->get_vars_size());
  memset(d->v, 0, d->get_vars_size());

  if (d->init_vars(d)) {
    return 1;
  }

  v->cl_opts = cli_create_option_array2(30);
  v->cl_args = cli_create_option_array2(5);
  /* Add std options & args */
  libdmn_add_opts(d, opts);
  libdmn_add_args(d, args);
  /* Add child options */
  if (d->add_opts)
    d->add_opts(d);
  /* Handle CLI args & opts */
  libdmn_handle_args(d, argc, argv);

  d->log = malloc(sizeof(log_desc_t));
  if ( d->log == NULL ) {
    return 1;
  }

  if ( log_init(d->log, &v->log_opts, v->name) ) {
    return 1;
  }
  d->h.dbg = d->log;

  DBG2("Initialization");

  /* Allocate buffer */
  v->buf = malloc(v->bufsize);
  if (!v->buf) {
    ERR2("Can't allocate memory");
    return 1;
  }

  v->fdl = fdlist_create(32);
  v->fdlr = fdlist_create(32);
  v->mng_fdl = fdlist_create(10);
  v->cmd = list_create(sizeof(libdmn_str_cmd_tbl_t), 50);
  v->fdhandlers = list_create(sizeof(libdmn_fd_handler_item_t), 20);
  v->timers = ptrlist_create(20);
  v->period_timers = ptrlist_create(20);

  if (d->sig_handler) {
    libdmn_setsighandler(d->log, SIGINT, d->sig_handler);
    libdmn_setsighandler(d->log, SIGTERM, d->sig_handler);
    libdmn_setsighandler(d->log, SIGHUP, d->sig_handler);
    libdmn_setsighandler(d->log, SIGALRM, d->sig_handler);
  }

  /* Create mng socket & add handlers */
  if (!v->cm[0]) {
   v->fd = socket(AF_UNIX, SOCK_STREAM, 0);
   if (v->fd == -1) {
     ERR2("Can't create mng socket: %s", strerror(errno));
     return 1;
   }

   unlink(v->sa.sun_path);

   if (bind(v->fd, (struct sockaddr *)&(v->sa), sizeof(v->sa))) {
     ERR2("Can't bind socket: %s", strerror(errno));
     return 1;
   }

   if (chmod(v->sa.sun_path, S_IWUSR | S_IWGRP | S_IWOTH | S_IRUSR | S_IRGRP | S_IROTH )) {
     ERR2("Can't change permissions for socket: %s", strerror(errno));
     return 1;
   }

   if (listen(v->fd, 10)) {
     ERR2("Can't listen socket: %s", strerror(errno));
     return 1;
   }

   /* Add std fd handlers */
   memset(&hit, 0, sizeof(hit));
   hit.init = libdmn_daemon_mng_init;
   hit.check = libdmn_daemon_mng_check;
   hit.handle_rx = libdmn_daemon_mng_handle_rx;
   hit.handle_close = libdmn_daemon_mng_handle_close;
   libdmn_add_fd_handler(d, &hit);

   memset(&hit, 0, sizeof(hit));
   hit.check = libdmn_daemon_mngcl_check;
   hit.handle_rx = libdmn_daemon_mngcl_handle_rx;
   hit.handle_close = libdmn_daemon_mngcl_handle_close;
   libdmn_add_fd_handler(d, &hit);

   /* Add std cmd handlers */
   libdmn_add_str_cmd_handlers(d, libdmn_str_cmd_table);

  }

  /* Execute command  & exit */
  if (v->cm[0]) {
    r  = libdmn_execute_cm(&d->h,
                           v->sa.sun_path, v->buf,
                           v->bufsize, v->cm);
    libdmn_daemon_exit(d, r);
  }

  if (d->init(d))
    return 1;

  if (!v->foreground)
    daemon(0, 0);

  /* save daemon startup time */
  time(&v->starttime);

  return 0;
}


int libdmn_handle_timeout(daemon_t *d)
{
  uint16_t i;
  int r = 0;
  uint64_t etime;
  daemon_common_var_t *v = &d->cv;
  libdmn_timer_t *timer;

  DBGL2(2, "");

  for (i=0; i<list_count(v->timers);++i) {

    timer = (libdmn_timer_t*)ptrlist_at(v->timers, i);
    etime = libdmn_gettime_ms()-timer->started;
    DBGL2(2, "t:%u p:%u et:%"PRIu64"", i, timer->period, etime);

    if (!timer->period)
      continue;

    if (etime>=timer->period) {
      timer->started = timer->started + (etime/timer->period)*timer->period;
      DBGL2(2, "nt:%"PRIu64, timer->started);
      if (timer->handler) {
        r = timer->handler(d, timer->data);
        if (r)
          ERR2("Timer %u handler failed: code %i", i, r);
      }
    }

  }
  return r;
}

int libdmn_timer_init(libdmn_timer_t *timer, libdmn_timer_handler_t handler, void *data)
{
  memset(timer, 0, sizeof(libdmn_timer_t));
  timer->data =data;
  timer->handler = handler; 

  return 0; 
}

int libdmn_timer_start(daemon_t *d, libdmn_timer_t *timer, uint32_t period)
{
  daemon_common_var_t *v = &d->cv;
  int i;

  if (period)
    timer->period = period;

  DBGL2(1,"p:%lu dt:%p", timer->period, timer->data);
  timer->started = libdmn_gettime_ms();

  i = ptrlist_find(v->timers, timer);
  if (i<0)
    ptrlist_add(v->timers, timer);

  return 0;
}

int libdmn_timer_stop(daemon_t *d, libdmn_timer_t *timer)
{
  daemon_common_var_t *v = &d->cv;
  ptrlist_del(v->timers, timer);
  return 0;
}

uint32_t libdmn_timer_elapsed(libdmn_timer_t *timer)
{
  uint32_t t;
  t = (libdmn_gettime_ms() - timer->started);
  if (timer->period < t )
     return 0;
  return timer->period - t;
}

void libdmn_add_periodical_handler(daemon_t *d, libdmn_timer_handler_t handler, uint16_t period)
{
  daemon_common_var_t *v = &d->cv;
  libdmn_timer_t *timer;
  timer = malloc(sizeof(libdmn_timer_t));
  libdmn_timer_init(timer, handler, d);
  libdmn_timer_start(d, timer, period);
  ptrlist_add(v->period_timers, timer);
}

int libdmn_handle_args(daemon_t *d, int argc, char **argv)
{
  daemon_common_var_t *v = &d->cv;
  int r=0;

  DBG2();
  r = cli_handle_opt2(v->cl_opts, v->cl_args, argc, argv);
  if (r)
   libdmn_daemon_exit(d, r);

  if (v->help) {      
    printf("\n%s - %s\n\n", v->name, v->description);
    if ((v->cl_args) && (v->cl_args[0])) {
      printf("Arguments:\n");
      cli_print_args_help2((const cli_option_item_t**)v->cl_args);
      printf("\n");
    }
    if ((v->cl_opts) && (v->cl_opts[0])) {
      printf("Options:\n");
      cli_print_options_help2((const cli_option_item_t**)v->cl_opts);
      printf("\n");
    }
    libdmn_daemon_exit(d, 0);
  }
  return r;
}


int libdmn_daemon_run(daemon_t *d, int argc, char **argv)
{
  daemon_common_var_t *v = &d->cv;
  int r, fd;
  libdmn_fd_handler_item_t *hitem;
  uint16_t i;

  /* Initialize common variables */
  libdmn_daemon_init_common_vars(d);

  /* Create & init daemon variables */
  if (libdmn_daemon_init(d, argc, argv)) {
    ERR2("Init failed");
    libdmn_daemon_exit(d, 5);
  }

  DBG2("init handlers");
  /* Init fd handlers */
  for (i=0; i < list_count(v->fdhandlers); ++i) {
    hitem = list_at(v->fdhandlers, i);
    if (hitem->init)
       hitem->init(d);
  }

  DBGL2(1,"fd %i ", fdlist_at(v->fdl, 0)->fd);
  /* Main daemon loop */
  while (1) {

    r = libdmn_wait_read_fds(d, v->fdl, v->fdlr, v->sleeptimer);        
    if (r==1) {
      libdmn_handle_timeout(d);
      continue;
    }

    DBGL2(2,"Wakeup fd count: %i ", fdlist_count(v->fdlr));
    for (i=0; i < fdlist_count(v->fdlr); ++i) {

      fd = fdlist_at(v->fdlr, i)->fd;
      DBGL2(2, "fd: %i - %i", i, fd);
      /* find event handler */
      hitem = find_fd_handler_item(d, fd);
      if (!hitem) {
        ERR2("Not handles fds %i", fd);
        break;
      }

      DBGL2(2,"found: %p - %p", hitem->handle_close, hitem->handle_rx);

      /* handle event */
      if (recv(fd, v->buf, 1, MSG_PEEK)!=1) {
        if (hitem->handle_close)
          hitem->handle_close(d, fd);
      } else {
        if (hitem->handle_rx)
          hitem->handle_rx(d, fd);
      }

    }

  }

  libdmn_daemon_exit(d, 0);
}
